"""Interfaces and implementations of different distributions for use in policies.

Most of the implementations here are thin wrappers around the appropriate `scipy` distributions used just to create
a common and more manageable interface for use within the rest of the library.
"""
